@--------------------------------------------------------------------------------------------------------
@ Fajl je napisan sa tabulacijom velicine 2 razmaka (podesiti editor radi preglednosti)
@ Komentari u ovom asembleru pocinju sa @ umjesto sa ;
@
@		Program prikazuje upotrebu virtuelne memorije kod ARM procesora, i to samo u varijanti
@		stranicenja u jednom nivou.	
@
@		Program sadrzi resetInterrupt od kojeg krece izvrsavanje kada se procesor ukljuci ili resetuje. Po 
@		pokretanju programa	prvo se inicijalizuju SP registri za modove koji ce se koristiti, osim za
@		user mod, sto se radi kasnije, pri prelasku u korisnicki mod. Potom se kopira sadrzaj
@		inicijalizovanih podataka (koji nisu konstantni) iz flash memorije u RAM memoriju, nakon cega se priprema
@		mehanizam za preslikavanje virtuelnih adresa u fizicke i pokrece korisnicki program.
@		Za potrebe aktiviranja mehanizma virtuelne memorije, prvo se kreira PMT tabela koja citav virtuelni adresni
@		prostor preslikava u fizicki adresni prostor 1-1 i to tako da se preslikavanjem virtuelne adrese dobije
@		fizicka adresa koja je ista kao virtuelna adresa. Za potrebe prikaza rada prekidne rutine za data miss,
@		jedan ulaz u PMT tabeli se proglasava nevazecim, ali se sadrzaj te stranice i dalje ostavlja u memoriji. 
@		Kada se pristupi toj stranici, u prekidnoj rutini se samo setuje ulaz u PMT. Posto se unaprijed znalo da
@		samo jedna stranica moze generisati data abort (zbog pocetnog sadrzaja PMT), u prekidnoj rutini se nije
@		citao sadrzaj registara koprocesora 15 koji sadrze virtuelnu adresu koja je izazvala data abort i razlog
@		zbog kojeg je doslo do abort-a (nevazeci ulaz u PMT ili prekrsaj prava pristupa).
@		Na kraju se adresa formirane PMT upise u MMU, pokrene mehanizam preslikavanja adresa, predje u 
@		korisnicki mod i potom pokrene korisnicki program (main funkcija iz "p3.c"). Po povratku iz korisnickog
@		programa, u beskonacnoj petlji se ceka gasenje ili reset procesora.
@
@--------------------------------------------------------------------------------------------------------

.global _Reset

.equ							RAM,		0xa0000000											@ Pocetna fizicka adresa RAM memorije.
.equ							PAGE_SIZE,	0x100000										@ Velicina stranice u bajtovima.

.equ							RAM_PAGE_0,	0xa00												@	Redni broj stranice cija je VA ista kao 
																													@	FA pocetka RAM.
																													@ Ovaj asembler radi sa oznacenim brojevima
																													@ pa to nije isto kao:
																													@ RAM / PAGE_SIZE.

.equ							STEP_BETWEEN_BLOCKS, 0x100000						@ Konstanta koju je potrebno dodati na
																													@ PMT ulaz kako bi se od ulaza koji se
																													@ preslikava u blok i dobio ulaz koji se
																													@ preslikava u blok i+1.

.equ							PMT_ENTRY_SIZE,	4												@ Velicina ulaza u PMT (u bajtovima).

.equ							OFFSET_OF_PMT_ENTRY_OF_RAM_PAGE_0,	(RAM_PAGE_0 * PMT_ENTRY_SIZE)	@ Pomjeraj od pocetka 
																													@ PMT do ulaza za stranu ciji je redni broj 
																													@ isti kao redni broj prvog bloka RAM memorije.

.equ							PMT_ENTRY_RAM_PAGE_0, (RAM | 0xc02)			@ Sadrzaj koji bi trebalo da ima ulaz da bi se 
																													@ preslikao u prvi blok RAM memorije.

.equ							PMT_ENTRY_CNT, 0x1000										@ Broj ulaza u PMT

.equ							WORD_SIZE,	4														@ Velicina rijeci u bajtovima

.equ							MASK_CLEAR_MODE,	0x1f									@ Maska za brisanje moda rada iz CPSR
.equ							MASK_USER_MODE,	0x10										@ Maska za postavljanje korisnickog moda.

.macro 						cpwait
									mrc 		p15, 0, r0, c2, c0, 0 					@ citanje proizvoljnog registra cp15
									mov 		r0, r0                					@ cekanje da se citanje koprocesora zavrsi
									sub 		pc, pc, #4            					@ skok na sledecu instrukciju
.endm

.text

@prekidne rutine

_Reset:						B 			resetInterrupt									@ Prekidna rutina za reset.
							B 			.
							B 			svcInterrupt										@ Prekidna rutina za swi (svc) - iz biblioteke stdla.a
							B 			.																@ Prekidna rutina za prefetch abort
							B 			dataAbortInterrupt							@ Prekidna rutina za data abort
							B 			.
							B 			.
							B 			.


@--------------------------------------------------------------------------------------------------------
@
@		Prekidna rutina za prekid data abort
@
@		Ova rutina samo prikazuje primjer pristupa i azuriranja PMT prilikom stranicne greske.
@		Stranica koja izaziva gresku je unaprijed poznata, pa se nije koristio koprocesor 15 za odredjivanje
@		iste. Stranica je vec bila ucitana u memoriju, pa je samo trebalo azurirati odgovarajuci ulaz u 
@		PMT tabeli.
@		Stranica koja je generisala gresku je stranica cija je virtuelna adresa ista kao fizicka adresa
@		prve stranice RAM memorije (stranica se ujedno i preslikava u isti blok). (pogledati kod za 
@		inicijalizovanje PMT koji pocinje od labele "nemaKopiranja").
@		RAM memorija u ovom sistemu pocinje od fizicke adrese 0xa0000000.
@		Dodatni biti (0xco2) koji se postavljaju u odgovarajuci ulaz PMT tabele definisu da ulaz opisuje
@		stranicu velicine 1MB, da je stranica zabranjena za kesiranje i baferisanje i da je dozvoljeno
@		citati iz i upisivati u datu stranicu.
@
@--------------------------------------------------------------------------------------------------------
.func							dataAbortInterrupt
dataAbortInterrupt:	push 	{r0, r1, r2, r3, r12, lr}
									ldr			r0, =PMT												@ Ucitava se adresa PMT.
									add 		r0, r0, #OFFSET_OF_PMT_ENTRY_OF_RAM_PAGE_0	@ Racuna se pomjeraj do ulaza koji odgovara prvoj stranici 
																													@ RAM (stranica 0xA00)
									ldr			r1, =PMT_ENTRY_RAM_PAGE_0				@	Sadrzaj PMT ulaza 0xa00, kada je data stranica u bloku 0xa00
									str 		r1, [r0]												@ PMT[RAM page 0] = 0xa0000c02 
	
									bl 			printInfoDataAbortInterrupt			@ Ispis poruke da se dogodio data abort prekid (p3.c)
									pop 		{r0, r1, r2, r3, r12, lr}
									subs		pc, lr, #8											@ iret
.endfunc



@--------------------------------------------------------------------------------------------------------
@
@		Prekidna rutina za prekid reset
@
@		Ovo je pocetna tacka izvrsavanja svaki put kada se sistem ukljuci ili resetuje. Kod na pocetku
@		inicijalizuje stekove (sadrzaje SP registara razlicitih modova) koji ce se koristiti kasnije u toku
@		rada (osim za korisnicki mod, koji se inicijalizuje tek po prelazu u korisnicki mod).
@		Program potom kopira pocetne vrijednosti iz FLASH memorije u RAM, zatim kreira jednu PMT koja
@		vrsi preslikavanje tako da je svaka virtuelna adresa jednaka fizickoj adresi u koju se preslikava.
@		Potom se inicijalizuje i pokrece MMU, prelazi u korisnicki mod, inicijalizuje stek u korisnickom modu
@		i prelazi na izvrsavanje main() funkcije iz fajla "p3.c". Kada se main zavrsi, ulazi se u beskonacnu
@		petlju i tu ostaje do gasenja ili resetovanja sistema.
@
@--------------------------------------------------------------------------------------------------------
	
.func							resetInterrupt
resetInterrupt:		ldr 		sp, =svcStackTop								@ Inicijalizovanje steka za supervizorski mod 
																													@ (svcStackTop inicijalizovano u "sc.ld").
																													
									mrs 		r0, cpsr												@ Prelaz u abort mod rada radi inicijalizovanja SP.
									ldr 		r1, =0x04												@ U r0 je ostao sadrzaj CPSR za potrebe povratka
									orr 		r1, r0, r1											@	u supervizorski mod. 
									msr 		cpsr, r1

									ldr			sp, =dataAbortStackTop					@ Inicijalizovanje steka za abort mod 
																													@ (svcStackTop inicijalizovano u "sc.ld").
																													
									msr			cpsr, r0												@ Povratak u supervizorski mod rada.
		
	
									ldr   	r0, =flashData									@	Adresa podataka u flesu sa koje se kopiraju
																													@	pocetne vrijednosti.
									ldr   	r1, =RAM												@ Adresa podataka u RAM na koju se kopiraju pocetne
																													@ vrijednosti.
									ldr   	r2, =dataSize										@	Broj bajtova koje treba iskopirati
									cmp   	r2, #0													@ (broj bajtova za kopiranje) > 0 ?
									ble 		nemaKopiranja										@	Ukoliko nema pocetnih vrijednosti koje treba kopirati
																													@ kopiranje se preskace.
									
copy:							ldr  		r4, [r0], #WORD_SIZE						@ Citanje jedne rijeci iz flash memorije 
									str  		r4, [r1], #WORD_SIZE						@ i upis te rijeci u RAM memoriju.
									subs  	r2, r2, #WORD_SIZE							@	Azurira se broj preostalih bajtova za kopiranje
									bgt			copy														@ if (broj preostalih bajtova za kopiranje) > 0)
																													@			then kopiraj jos (goto copy).

																													@ Popunjavanje PMT tabele:						
nemaKopiranja:		ldr 		r0, =PMT												@	Adresa PMT
									mov 		r1, #PMT_ENTRY_CNT							@	Broj ulaza u PMT
									ldr 		r2, =0x00c02										@	Biti koje treba setovati u svim ulazima.
																													@ Za ovaj primjer je to korektno.
																													@ U opstem slucaju mogu i treba da se razlikuju.
																													@ To je ujedno i sadrzaj PMT ulaza za stranicu koja se
																													@ preslikava u nulti blok.
									mov 		r3, #STEP_BETWEEN_BLOCKS				@ Razlika izmedju dva susjedna bloka, posmatrano iz ugla
																													@ sadrzaja PMT ulaza.
									
pmtFill:					str 		r2, [r0],#4											@ Stranica i se preslikava u blok i.
									add 		r2, r2, r3											@ Uvecavanje rednog broja bloka - priprema sadrzaja za
																													@ sledeci ulaz u PMT.
									subs 		r1, r1, #1											@ (broj ulaza u PMT koje treba popuniti)++
									bne 		pmtFill													@ if (broj ulaza u PMT koje treba popuniti) != 0 
																													@		then popnjavaj dalje
	
																																			@ Brisanje jednog ulaza u PMT za potrebe
																																			@ prikaza rada prekidne rutine za data abort.	
									ldr 		r0, =PMT																		@ Ucitava se adresa PMT.
									add 		r0, r0, #OFFSET_OF_PMT_ENTRY_OF_RAM_PAGE_0	@ Izracunava se adresa stranice koja se
																																			@ preslikava u prvi blok RAM memorije,
									mov 		r1, #0
									str 		r1, [r0]																		@ i za istu se zapisuje da nije u memoriji.
																																			@ Sadrzaj stranice i dalje ostaje u memoriji,
																																			@	ali ce pri pokusaju pristupa stranici doci
																																			@ do stranicne greske. Zato je u prekidnoj
																																			@ rutini dovoljno postaviti validan sadrzaj
																																			@ ulaza, pa ce se strana moci koristiti.

									ldr 		r0, =PMT												
									mcr 		p15,0,r0,c2,c0,0								@	Upisivanje adrese pocetka PMT u registar 2 cp15.
	
									ldr 		r0, =0x1
									mcr 		p15,0,r0,c3,c0,0								@ Upis vrijednosti u registar za kontrolu pristupa domenima.
																													@	Sve stranice su u klijentskom modu:
																													@		-supervizor ima puno pravo pristupa;
																													@		-korisnik dobija prava na osnovu sadrzaja AP bita PMT ulaza.

									mrc 		p15,0,r0,c1,c0,0   							@ Citanje registra 1 cp15 radi setovanja poslednjeg bita 
																													@ (ukljucivanja MMU).
									orr 		r0, r0, #0x1      							@ Setovanje poslednjeg bita u procitanom sadrzaju.
									mcr 		p15,0,r0,c1,c0,0   							@ Vracanje sadrzaja u registar 1 cp15 sa setovanim 
																													@ enable bitom za MMU.

									cpwait																	@ Makro za cekanje zavrsetka svih upisa u MMU 
																													@ (dofinisan na pocetku fajla).

																													@ Prelazak u korisnicki mod rada
									mrs 		r0, cpsr												@ Ucitavanje cpsr radi zamjene poslednjih 5 bita (mod rada).
									ldr 		r1, =MASK_CLEAR_MODE
									bic 		r0, r0, r1											@ Brisanje poslednjih 5 bita dosadasnjeg moda rada.
									ldr 		r1, =MASK_USER_MODE
									orr 		r0, r0, r1											@ Setovanje poslednjih 5 bita za korisnicki mod rada.
									msr 		cpsr, r0												@ Upis nove vrijednosti u CPSR sa izmjenjenim modom rada.
									
									ldr 		sp, =userStackTop								@ Inicijalizovanje steka za korisnicki mod rada.
									
 									bl 			main														@ Pokretanje programa iz (p3.c)
									b 			.																@ Beskonacna petlja u kojoj procesor ceka reset ili gasenje.
.endfunc
 
.end
 
